# coding: utf-8
"""
These functions help you use hydrology analysis.
"""

import logging as _logging
import arcgis
from arcgis.geoprocessing._support import _execute_gp_tool
from arcgis.features import FeatureSet
from arcgis.features.geo._accessor import _is_geoenabled
from pandas import DataFrame

_log = _logging.getLogger(__name__)


def _evaluate_spatial_input(input_points):
    """
    Helper function to determine if the input is either a FeatureSet or Spatially Enabled DataFrame, and
    output to FeatureSet for subsequent processing.
    :param input_points: FeatureSet or Spatially Enabled DataFrame
    :return: FeatureSet
    """
    if isinstance(FeatureSet):
        return FeatureSet

    elif isinstance(input_points, DataFrame) and _is_geoenabled(input_points):
        return input_points.spatial.to_featureset()

    elif isinstance(input_points, DataFrame) and not _is_geoenabled(input_points):
        raise Exception('input_points is a DataFrame, but does not appear to be spatially enabled. Using the <df>.spatial.set_geometry(col, sr=None) may help. (https://esri.github.io/arcgis-python-api/apidoc/html/arcgis.features.toc.html#arcgis.features.GeoAccessor.set_geometry)')

    else:
        raise Exception('input_points must be either a FeatureSet or Spatially Enabled DataFrame instead of {}'.format(type(input_points)))


def trace_downstream(input_points, point_id_field=None, source_database='Finest', generalize=False,
                     gis=None):
    """
    The Trace Downstream method delineates the downstream path from a specified location. 
    Esri-curated elevation data is used to create an output polyline delineating the flow path 
    downstream from the specified input location. This method accesses a service using multiple 
    source databases which are available for different geographic areas and at different 
    spatial scales.
    
    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    input_points           Required Feature Set or Spatially Enabled DataFrame
                           Points delineating the starting location to calculate the downstream 
                           location from.
    ------------------     --------------------------------------------------------------------
    point_id_field         Optional String
                           Field used to identify the feature from the source data. This is 
                           useful for relating the results back to the original source data.
    ------------------     --------------------------------------------------------------------
    source_database        Optional String - Default "Finest"
                           Keyword indicating the source data that will be used in the
                           analysis. This keyword is an approximation of the spatial resolution
                           of the digital elevation model used to build the foundation
                           hydrologic  database. Since many elevation sources are distributed
                           with units of  arc seconds, this keyword is an approximation in
                           meters for easier understanding.


                           - Finest: Finest resolution available at each location from all
                             possible data sources.
                           - 10m: The hydrologic source was built from 1/3 arc second -
                             approximately 10 meter resolution, elevation data.
                           - 30m: The hydrologic source was built from 1 arc second -
                             approximately 30 meter resolution, elevation data.
                           - 90m: The hydrologic source was built from 3 arc second -
                             approximately 90 meter resolution, elevation data.
    ------------------     --------------------------------------------------------------------
    generalize             Optional Boolean - Default False
                           Determines if the output downstream trace lines will be smoothed 
                           into simpler lines.
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS Object instance
                           If not provided as input, a GIS object instance logged into an 
                           active portal with elevation helper services defined must already 
                           be created in the active Python session. A GIS object instance can 
                           also be optionally explicitly passed in through this parameter.
    ==================     ====================================================================

    :return:
       FeatureSet
    """
    kwargs = locals()

    param_db = {
        "input_points": (FeatureSet, "InputPoints"),
        "point_id_field": (str, 'PointIDField'),
        "source_database": (str, 'SourceDatabase'),
        "generalize": (str, 'Generalize'),
        "output_trace_line": (FeatureSet, "Output Trace Line"),
    }

    return_values = [
        {"name": "output_trace_line", "display_name": "Output Trace Line", "type": FeatureSet},
    ]

    # use helper function to evaluate the input points and convert them, if necessary, to a FeatureSet
    input_fs = _evaluate_spatial_input(input_points)

    if input_fs.geometry_type != 'esriGeometryPoint':
        raise Exception('input_points FeatureSet must be point esriGeometryPoint, not {}'.format(input_fs.geometry_type))

    input_fields = input_fs.fields
    if point_id_field and point_id_field not in [f['name'] for f in input_fields] and len(input_fields):
        input_fields_str = ','.join(input_fields)
        raise Exception('The provided point_id_field {} does not appear to be in the input_points FeatureSet fields - {}'.format(point_id_field, input_fields_str))

    if source_database not in ['Finest', '10m', '30m', '90m']:
        raise Exception('source_database must be either "Finest", "10m", "30m", or "90m". {} does not appear to be one of these.'.format(source_database))

    if gis is None and arcgis.env.active_gis is None:
        raise Exception('GIS must be defined either by directly passing in a GIS object created using credentials, or one must already be created in the active Python session.')
    elif gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.hydrology.url

    return _execute_gp_tool(gis, "TraceDownstream", kwargs, param_db, return_values, True, url)


def watershed(input_points, point_id_field=None, snap_distance=10, snap_distance_units='Meters',
              source_database='Finest', generalize=False, gis=None):
    """
    The Watershed task is used to identify catchment areas based on a particular location you
    provide and ArcGIS Online Elevation data.


    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    input_points           Required Feature Set or Spatially Enabled DataFrame
                           Points delineating the starting location to calculate the downstream
                           location from.
    ------------------     --------------------------------------------------------------------
    point_id_field         Optional String
                           Field used to identify the feature from the source data. This is
                           useful for relating the results back to the original source data.
    ------------------     --------------------------------------------------------------------
    snap_distance          Optional Integer - Default 10
    ------------------     --------------------------------------------------------------------
    snap_distance_units    Optional String - Default Meters
                           Meters | Kilometers | Feet | Yards | Miles
    ------------------     --------------------------------------------------------------------
    source_database        Optional String - Default "Finest"
                           Keyword indicating the source data that will be used in the analysis.
                           This keyword is an approximation of the spatial resolution of the
                           digital elevation model used to build the foundation hydrologic
                           database. Since many elevation sources are distributed with units of
                           arc seconds, this keyword is an approximation in meters for easier
                           understanding.

                           - Finest: Finest resolution available at each location from all
                             possible data sources.
                           - 10m: The hydrologic source was built from 1/3 arc second -
                             approximately 10 meter resolution, elevation data.
                           - 30m: The hydrologic source was built from 1 arc second -
                             approximately 30 meter resolution, elevation data.
                           - 90m: The hydrologic source was built from 3 arc second -
                             approximately 90 meter resolution, elevation data.
    ------------------     --------------------------------------------------------------------
    generalize             Optional Boolean - Default False
                           Determines if the output downstream trace lines will be smoothed
                           into simpler lines.
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS Object instance
                           If not provided as input, a GIS object instance logged into an
                           active portal with elevation helper services defined must already
                           be created in the active Python session. A GIS object instance can
                           also be optionally explicitly passed in through this parameter.
    ==================     ====================================================================

    :return:
        Result object comprised of two FeatureSets - one for watershed_area, and another for
        snapped_points
    """

    kwargs = locals()

    param_db = {
        "input_points": (FeatureSet, "InputPoints"),
        "point_id_field": (str, 'PointIDField'),
        "snap_distance": (int, 'SnapDistance'),
        "snap_distance_units": (int, 'SnapDistanceUnits'),
        "source_database": (str, 'SourceDatabase'),
        "generalize": (str, 'Generalize'),
        "watershed_area": (FeatureSet, "Watershed Area"),
        "snapped_points": (FeatureSet, "SnappedPoints")
    }

    return_values = [
        {"name": "watershed_area", "display_name": "Watershed Area", "type": FeatureSet},
        {"name": "snapped_points", "display_name": "Snapped Points", "type": FeatureSet}
    ]

    # use helper function to evaluate the input points and convert them, if necessary, to a FeatureSet
    input_fs = _evaluate_spatial_input(input_points)

    if input_fs.geometry_type != 'esriGeometryPoint':
        raise Exception('input_points FeatureSet must be point esriGeometryPoint, not {}.'.format(input_fs.geometry_type))

    input_fields = input_fs.fields
    if point_id_field and point_id_field not in [f['name'] for f in input_fields] and len(input_fields):
        input_fields_str = ','.join(input_fields)
        raise Exception('The provided point_id_field {} does not appear to be in the input_points FeatureSet fields - {}'.format(point_id_field, input_fields_str))

    if gis is None and arcgis.env.active_gis is None:
        raise Exception('GIS must be defined either by directly passing in a GIS object created using credentials, or one must already be created in the active Python session.')
    elif gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.hydrology.url

    return _execute_gp_tool(gis, "Watershed", kwargs, param_db, return_values, True, url)



from fastai.vision import ImageSegment, Image
from fastai.vision.image import open_image, show_image, pil2tensor
from fastai.vision.data import SegmentationProcessor, ImageItemList
from fastai.layers import CrossEntropyFlat
from fastai.basic_train import LearnerCallback
import warnings
import PIL
import numpy as np

class ArcGISImageSegment(Image):
    "Support applying transforms to segmentation masks data in `px`."
    def __init__(self, x, cmap=None, norm=None):
        super(ArcGISImageSegment, self).__init__(x)
        self.cmap = cmap
        self.mplnorm = norm

    def lighting(self, func, *args, **kwargs):
        return self

    def refresh(self):
        self.sample_kwargs['mode'] = 'nearest'
        return super().refresh()

    @property
    def data(self):
        "Return this image pixels as a `LongTensor`."
        return self.px.long()

    def show(self, ax=None, figsize:tuple=(3,3), title=None, hide_axis:bool=True,
        cmap='tab20', alpha:float=0.5, **kwargs):
        "Show the `ImageSegment` on `ax`."
        ax = show_image(self, ax=ax, hide_axis=hide_axis, cmap=self.cmap, figsize=figsize,
                        interpolation='nearest', alpha=alpha, vmin=0, norm=self.mplnorm, **kwargs)
        if title: ax.set_title(title)

def is_no_color(color_mapping):
    if isinstance(color_mapping, dict):
        color_mapping = list(color_mapping.values())
    return (np.array(color_mapping) == [-1., -1., -1.]).any()

class ArcGISSegmentationLabelList(ImageItemList):
    "`ItemList` for segmentation masks."
    _processor = SegmentationProcessor
    def __init__(self, items, classes=None, class_mapping=None, color_mapping=None, **kwargs):
        super().__init__(items, **kwargs)
        self.class_mapping = class_mapping
        self.color_mapping = color_mapping
        self.copy_new.append('classes')
        self.classes, self.loss_func = classes, CrossEntropyFlat(axis=1)
        if is_no_color(list(color_mapping.values())):
            self.cmap = 'tab20'  ## compute cmap from palette
            import matplotlib as mpl
            bounds = list(color_mapping.keys())
            self.mplnorm = mpl.colors.BoundaryNorm(bounds, len(bounds))
        else:
            import matplotlib as mpl
            self.cmap = mpl.colors.ListedColormap(np.array(list(color_mapping.values()))/255)
            bounds = list(color_mapping.keys())
            self.mplnorm = mpl.colors.BoundaryNorm(bounds, self.cmap.N)

    def open(self, fn):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning) # EXIF warning from TiffPlugin
            x = PIL.Image.open(fn)
            if x.palette is not None:
                x = x.convert('P')
            else:
                x = x.convert('L')
            x = pil2tensor(x, np.float32)

        return ArcGISImageSegment(x, cmap=self.cmap, norm=self.mplnorm)

    def analyze_pred(self, pred, thresh:float=0.5): 
        label_mapping = {(idx + 1):value for idx, value in enumerate(self.class_mapping.keys())}
        out = pred.argmax(dim=0)[None]
        for key, value in label_mapping.items():
            out[out==key] = value
        return out

    def reconstruct(self, t): 
        return ArcGISImageSegment(t, cmap=self.cmap, norm=self.mplnorm)

class ArcGISSegmentationItemList(ImageItemList):
    "`ItemList` suitable for segmentation tasks."
    _label_cls, _square_show_res = ArcGISSegmentationLabelList, False

class LabelCallback(LearnerCallback):
    def __init__(self, learn):
        super().__init__(learn)
        import pdb
        self.label_mapping = {value:(idx+1) for idx, value in enumerate(learn.data.class_mapping.keys())}
        
    def on_batch_begin(self, last_input, last_target, **kwargs):
        for label, idx in self.label_mapping.items():
            last_target[last_target==label] = idx
        return last_input, last_target


"""
These functions help you answer one of the most common questions posed in spatial analysis: "What is near what?"

connect_origins_to_destinations measures the travel time or distance between pairs of points.
create_buffers create areas of equal distance from features.
create_drive_time_areas finds areas around locations that can be reached within a time period.
find_nearest identifies those places that are the closest to known locations.
plan_routes determines the best way to route a fleet of vehicles to visit many stops.
"""
import arcgis as _arcgis
from arcgis._impl.common._utils import _date_handler

def connect_origins_to_destinations(origins_layer,
                                    destinations_layer,
                                    measurement_type="DrivingTime",  
                                    origins_layer_route_id_field=None,
                                    destinations_layer_route_id_field=None,
                                    time_of_day=None,
                                    time_zone_for_time_of_day="GeoLocal",
                                    output_name=None,
                                    context=None,
                                    gis=None,
                                    estimate=False,
                                    point_barrier_layer=None,
                                    line_barrier_layer=None,
                                    polygon_barrier_layer=None):
    """
    The Connect Origins to Destinations task measures the travel time or distance between pairs of points. Using this tool, you can

    * Calculate the total distance or time commuters travel on their home-to-work trips.
    * Measure how far customers are traveling to shop at your stores. Use this information to define your market reach, especially when targeting advertising campaigns or choosing new store locations.
    * Calculate the expected trip mileage for your fleet of vehicles. Afterward, run the Summarize Within tool to report mileage by state or other region.

    You provide starting and ending points, and the tool returns a layer containing route lines, including measurements, between the 
    paired origins and destinations.

    ===================================    =========================================================
    **Argument**                           **Description**
    -----------------------------------    ---------------------------------------------------------
    origins_layer                          Required layer. The starting point or points of the routes to be generated. See :ref:`Feature Input<FeatureInput>`.
    -----------------------------------    ---------------------------------------------------------
    destinations_layer                     Required layer. The routes end at points in the destinations layer. See :ref:`Feature Input<FeatureInput>`.
    -----------------------------------    ---------------------------------------------------------
    measurement_type                       Required string. The origins and destinations can be connected by measuring straight-line distance, 
                                           or by measuring travel time or travel distance along a street network using various modes of transportation 
                                           known as travel modes.
                                           
                                           Valid values are a string, StraightLine, which indicates Euclidean distance to be used as distance measure or 
                                           a Python dictionary representing settings for a travel mode.
                                           
                                           When using a travel mode for the measurement_type, you need to specify a dictionary 
                                           containing the settings for a travel mode supported by your organization. The code in the example section below generates 
                                           a valid Python dictionary and then passes it as the value for the measurement_type parameter.

                                           Supported travel modes: ['Driving Distance', 'Driving Time', 'Rural Driving Distance', 'Rural Driving Time', 
                                           'Trucking Distance', 'Trucking Time', 'Walking Distance', 'Walking Time']
                                           
    -----------------------------------    ---------------------------------------------------------
    origins_layer_route_id_field           Optional string. Specify the field in the origins layer containing the IDs that pair origins with destinations.

                                           * The ID values must uniquely identify points in the origins layer.

                                           * Each ID value must also correspond with exactly one route ID value in the destinations layer. Route IDs that match 
                                             across the layers create origin-destination pairs, which the tool connects together.

                                           *  Specifying origins_layer_route_id_field is optional when there is exactly one point feature in the origins or 
                                              destinations layer. The tool will connect all origins to the one destination or the one origin to all destinations, 
                                              depending on which layer contains one point.
    -----------------------------------    ---------------------------------------------------------
    destinations_layer_route_id_field      Optional string. Specify the field in the destinations layer containing the IDs that pair origins with destinations.

                                           * The ID values must uniquely identify points in the destinations layer.

                                           * Each ID value must also correspond with exactly one route ID value in the origins layer. Route IDs that match across the 
                                             layers create origin-destination pairs, which the tool connects together.

                                           * Specifying destinations_layer_route_id_field is optional when there is exactly one point feature in the origins or 
                                             destinations layer. The tool will connect all origins to the one destination or the one origin to all destinations, 
                                             depending on which layer contains one point.
    -----------------------------------    ---------------------------------------------------------
    time_of_day                            Optional datetime.datetime. Specify whether travel times should consider traffic conditions. To use traffic in the analysis, 
                                           set measurement_type to a travel mode object whose impedance_attribute_name property is set to travel_time and assign a value 
                                           to time_of_day. (A travel mode with other impedance_attribute_name values don't support traffic.) The time_of_day value represents 
                                           the time at which travel begins, or departs, from the origin points. The time is specified as datetime.datetime.

                                           The service supports two kinds of traffic: typical and live. Typical traffic references travel speeds that are made up of historical 
                                           averages for each five-minute interval spanning a week. Live traffic retrieves speeds from a traffic feed that processes phone probe 
                                           records, sensors, and other data sources to record actual travel speeds and predict speeds for the near future.
                                           
                                           The `data coverage <http://www.arcgis.com/home/webmap/viewer.html?webmap=b7a893e8e1e04311bd925ea25cb8d7c7>`_ page shows the countries 
                                           Esri currently provides traffic data for.
                                           
                                           Typical Traffic:

                                           To ensure the task uses typical traffic in locations where it is available, choose a time and day of the week, and then convert the day 
                                           of the week to one of the following dates from 1990:

                                           * Monday—1/1/1990
                                           * Tuesday—1/2/1990
                                           * Wednesday—1/3/1990
                                           * Thursday—1/4/1990
                                           * Friday—1/5/1990
                                           * Saturday—1/6/1990
                                           * Sunday—1/7/1990
                                           Set the time and date as datetime.datetime.

                                           For example, to solve for 1:03 p.m. on Thursdays, set the time and date to 1:03 p.m., 4 January 1990; and convert to 
                                           datetime eg. datetime.datetime(1990, 1, 4, 1, 3).
                                           
                                           Live Traffic:

                                           To use live traffic when and where it is available, choose a time and date and convert to datetime.

                                           Esri saves live traffic data for 12 hours and references predictive data extending 12 hours into the future. If the time and date you 
                                           specify for this parameter is outside the 24-hour time window, or the travel time in the analysis continues past the predictive data window, the task falls back to typical traffic speeds.
                                           
                                           Examples:
                                           from datetime import datetime

                                           * "time_of_day": datetime(1990, 1, 4, 1, 3) # 13:03, 4 January 1990. Typical traffic on Thursdays at 1:03 p.m.
                                           * "time_of_day": datetime(1990, 1, 7, 17, 0) # 17:00, 7 January 1990. Typical traffic on Sundays at 5:00 p.m.
                                           * "time_of_day": datetime(2014, 10, 22, 8, 0) # 8:00, 22 October 2014. If the current time is between 8:00 p.m., 21 Oct. 2014 and 8:00 p.m., 22 Oct. 2014, 
                                             live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.
                                           * "time_of_day": datetime(2015, 3, 18, 10, 20) # 10:20, 18 March 2015. If the current time is between 10:20 p.m., 17 Mar. 2015 and 10:20 p.m., 18 Mar. 2015, 
                                             live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.

    -----------------------------------    ---------------------------------------------------------
    time_zone_for_time_of_day              Optional string. Specify the time zone or zones of the timeOfDay parameter. 
                                           Choice list: ['GeoLocal', 'UTC']
                                           
                                           GeoLocal-refers to the time zone in which the originsLayer points are located.
                                           
                                           UTC-refers to Coordinated Universal Time.
    -----------------------------------    ---------------------------------------------------------
    include_route_layers                   Optional Boolean. When include_route_layers is set to true, each route from the result is also saved as a route layer item. A route layer includes all the 
                                           information for a particular route such as the stops assigned to the route as well as the travel directions. Creating route layers is useful if you want 
                                           to share individual routes with other members in your organization. The route layers use the output feature service name provided in the outputName 
                                           parameter as a prefix and the route name generated as part of the analysis is added to create a unique name for each route layer.
                                           
                                           Caution: Route layers cannot be created when the output is a feature collection. The task will raise an error if output_name is not specified 
                                           (which indicates feature collection output) and include_route_layers is true.

                                           The maximum number of route layers that can be created is 1,000. If the result contains more than 1,000 routes and include_route_layers is true, 
                                           the task will only create the output feature service.
    -----------------------------------    ---------------------------------------------------------          
    output_name                            Optional string. If provided, the task will create a feature layer of the results. You define the name of the layer. If output_name is not supplied, 
                                           the task will return a feature collection.
    -----------------------------------    ---------------------------------------------------------
    context                                Optional string. Additional settings such as processing extent and output spatial reference. For calculate_density, there are two settings.

                                           #. Extent (extent)-a bounding box that defines the analysis area. Only those points in the origins_layer and destinations_layer that intersect the 
                                              bounding box will be analyzed.	
                                           #. Output Spatial Reference (outSR)

                                           * If the output is a feature service, the spatial reference will be the same as originsLayer. Setting outSR for feature services has no effect.
                                           
                                           * If the output is a feature collection, the features will be in the spatial reference of the outSRvalue or the spatial reference of originsLayer 
                                             when outSR is not specified.
    -----------------------------------    ---------------------------------------------------------
    gis                                    Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -----------------------------------    ---------------------------------------------------------
    estimate                               Optional Boolean. Is true, the number of credits needed to run the operation will be returned as a float.
    -----------------------------------    ---------------------------------------------------------   
    point_barrier_layer                    Optional layer. Specify one or more point features that act as temporary restrictions (in other words, barriers) when traveling on the underlying streets.

                                           A point barrier can model a fallen tree, an accident, a downed electrical line, or anything that completely blocks traffic at a specific position along 
                                           the street. Travel is permitted on the street but not through the barrier. See :ref:`Feature Input<FeatureInput>`.
    -----------------------------------    ---------------------------------------------------------
    line_barrier_layer                     Optional layer. Specify one or more line features that prohibit travel anywhere the lines intersect the streets.

                                           A line barrier prohibits travel anywhere the barrier intersects the streets. For example, a parade or protest that blocks traffic across several street 
                                           segments can be modeled with a line barrier. See :ref:`Feature Input<FeatureInput>`.
    -----------------------------------    ---------------------------------------------------------
    polygon_barrier_layer                  Optional string. Specify one or more polygon features that completely restrict travel on the streets intersected by the polygons.

                                           One use of this type of barrier is to model floods covering areas of the street network and making road travel there impossible. See :ref:`Feature Input<FeatureInput>`. 
    ===================================    =========================================================


    :returns: dict with the following keys:
        
        "routes_layer" : layer (FeatureCollection)
        
        "unassigned_origins_layer" : layer (FeatureCollection)
        
        "unassigned_destinations_layer" : layer (FeatureCollection)
    .. code-block:: python

        USAGE EXAMPLE: To retrieve trvel modes and run connect_origins_to_destinations tool. 
        
        This example creates route between esri regional offices to esri headquarter.
        	
        import arcgis.network as network
        route_service = network.RouteLayer(gis.properties.helperServices.route.url, gis=gis)
        travel_mode = [i for i in route_service.retrieve_travel_modes()['supportedTravelModes'] 
            if i['name'] == 'Rural Driving Distance'][0]
        routes =  connect_origins_to_destinations(origins_layer=esri_regional,
                                         destinations_layer=dest_layer,
                                         measurement_type=travel_mode,
                                         time_of_day=datetime(1990, 1, 4, 1, 3),
                                         output_name="routes_from_offices_to_hq")   
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.connect_origins_to_destinations(
        origins_layer,
        destinations_layer,
        measurement_type,
        origins_layer_route_id_field,
        destinations_layer_route_id_field,
        _date_handler(time_of_day),
        time_zone_for_time_of_day,
        output_name,
        context,
        estimate=estimate)


def create_buffers(
        input_layer,
        distances=[],
        field=None,
        units="Meters",
        dissolve_type="None",
        ring_type="Disks",
        side_type="Full",
        end_type="Round",
        output_name=None,
        context=None,
        gis=None,
        estimate=False):
    """
    Creates buffer polygon(s) around input features.

    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)
        The input to be buffered.
    distances : Optional list of floats
        The distance(s) that will be buffered.
    field : Optional string
        Buffers will be created using field values.
    units : Optional string
        The linear unit to be used with the distance value(s).
    dissolve_type : Optional string
        Specifies the dissolve to be performed to remove buffer overlap.
    ring_type : Optional string
        The ring type.
    side_type : Optional string
        The side(s) of the input that will be buffered.
    end_type : Optional string
        The shape of the buffer at the end of buffered line features.
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    estimate :
        Optional Boolean. If True, the number of credits to run the operation will be returned.

    Returns
    -------
    buffer_layer : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.create_buffers(input_layer,
                                                     distances,
                                                     field,
                                                     units,
                                                     dissolve_type,
                                                     ring_type,
                                                     side_type,
                                                     end_type,
                                                     output_name,
                                                     context,
                                                     estimate=estimate)


def create_drive_time_areas(input_layer,
                            break_values=[5, 10, 15],
                            break_units="Minutes",
                            travel_mode="Driving",
                            overlap_policy="Overlap",
                            time_of_day=None,
                            time_zone_for_time_of_day="GeoLocal",
                            output_name=None,
                            context=None,
                            gis=None,
                            estimate=False):
    """


    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)

    break_values : Optional list of floats

    break_units : Optional string

    travel_mode : Optional string

    overlap_policy : Optional string

    time_of_day : Optional datetime.datetime

    time_zone_for_time_of_day : Optional string

    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    estimate :
        Optional Boolean. If True, the number of credits to run the operation will be returned.

    Returns
    -------
    drive_time_areas_layer : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.create_drive_time_areas(
        input_layer,
        break_values,
        break_units,
        travel_mode,
        overlap_policy,
        _date_handler(time_of_day),
        time_zone_for_time_of_day,
        output_name,
        context,
        estimate=estimate)


def find_nearest(
        analysis_layer,
        near_layer,
        measurement_type="StraightLine",
        max_count=100,
        search_cutoff=2147483647,
        search_cutoff_units=None,
        time_of_day=None,
        time_zone_for_time_of_day="GeoLocal",
        output_name=None,
        context=None,
        gis=None,
        estimate=False):
    """
    Measures the straight-line distance, driving distance, or driving time from features in the analysis layer to
    features in the near layer, and copies the nearest features in the near layer to a new layer. Returns a layer
    containing the nearest features and a line layer that links the start locations to their nearest locations.

    Parameters
    ----------
    analysis_layer : Required layer (see Feature Input in documentation)
        For each feature in this layer, the task finds the nearest features from the nearLayer.
    near_layer : Required layer (see Feature Input in documentation)
        The features from which the nearest locations are found.
    measurement_type : Required string
        The nearest locations can be determined by measuring straight-line distance, driving distance, or driving time
    max_count : Optional int
        The maximum number of near locations to find for each feature in analysisLayer.
    search_cutoff : Optional float
        Limits the search range to this value
    search_cutoff_units : Optional string
        The units for the value specified as searchCutoff
    time_of_day : Optional datetime.datetime
        When measurementType is DrivingTime, this value specifies the time of day to be used for driving time
        calculations based on traffic.
    time_zone_for_time_of_day : Optional string

    output_name : Optional string
        Additional properties such as output feature service name
    context : Optional string
        Additional settings such as processing extent and output spatial reference
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    estimate :
        Optional Boolean. If True, the number of credits to run the operation will be returned.

    Returns
    -------
    dict with the following keys:
       "nearest_layer" : layer (FeatureCollection)
       "connecting_lines_layer" : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_nearest(
        analysis_layer,
        near_layer,
        measurement_type,
        max_count,
        search_cutoff,
        search_cutoff_units,
        _date_handler(time_of_day),
        time_zone_for_time_of_day,
        output_name,
        context,
        estimate=estimate)


def plan_routes(
        stops_layer,
        route_count,
        max_stops_per_route,
        route_start_time,
        start_layer,
        start_layer_route_id_field=None,
        return_to_start=True,
        end_layer=None,
        end_layer_route_id_field=None,
        travel_mode="Driving",
        stop_service_time=0,
        max_route_time=525600,
        include_route_layers=False,
        output_name=None,
        context=None,
        gis=None,
        estimate=False):
    """
    You provide a set of stops and the number of vehicles available to visit the stops, and Plan Routes determines how
    to efficiently assign the stops to the vehicles and route the vehicles to the stops.

    Use this tool to plan work for a mobile team of inspectors, appraisers, in-home support service providers, and
    others; deliver or pick up items from remote locations; or offer transportation services to people.

    Parameters
    ----------

    stops_layer : Required layer (see Feature Input in documentation)

    route_count : Required int

    max_stops_per_route : Required int

    route_start_time : Required datetime.datetime

    start_layer : Required layer (see Feature Input in documentation)

    start_layer_route_id_field : Optional string

    return_to_start : Optional bool

    end_layer : Optional layer (see Feature Input in documentation)

    end_layer_route_id_field : Optional string

    travel_mode : Optional string

    stop_service_time : Optional float

    max_route_time : Optional float

    include_route_layers : Optional bool

    output_name : Optional string

    context : Optional string

    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    estimate :
        Optional Boolean. If True, the number of credits to run the operation will be returned.

    Returns
    -------
    dict with the following keys:
       "routes_layer" : layer (FeatureCollection)
       "assigned_stops_layer" : layer (FeatureCollection)
       "unassigned_stops_layer" : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.plan_routes(
        stops_layer,
        route_count,
        max_stops_per_route,
        _date_handler(route_start_time),
        start_layer,
        start_layer_route_id_field,
        return_to_start,
        end_layer,
        end_layer_route_id_field,
        travel_mode,
        stop_service_time,
        max_route_time,
        include_route_layers,
        output_name,
        context,
        estimate=estimate)



"""
These functions help you identify, quantify, and visualize spatial patterns in your data.

calculate_density takes known quantities of some phenomenon and spreads these quantities across the map.
find_hot_spots identifies statistically significant clustering in the spatial pattern of your data.
interpolate_points predicts values at new locations based on measurements found in a collection of points.
"""

import arcgis as _arcgis

def calculate_density(
        input_layer,
        field=None,
        cell_size=None,
        cell_size_units="Meters",
        radius=None,
        radius_units=None,
        bounding_polygon_layer=None,
        area_units=None,
        classification_type="EqualInterval",
        num_classes=10,
        output_name=None,
        context=None,
        gis=None,
        estimate=False):
    """
    The calculate_density function creates a density map from point or line features by spreading known quantities of
    some phenomenon (represented as attributes of the points or lines) across the map. The result is a layer of areas
    classified from least dense to most dense.

    For point input, each point should represent the location of some event	or incident, and the result layer represents
    a count of the incident per unit area. A higher density value in a new location means that there are more points near
    that location. In many cases, the result layer can be interpreted as a risk surface for future events. For example,
    if the input points represent locations of lightning strikes, the result layer can be interpreted as a risk surface
    for future lightning strikes.	
            
    For line input, the line density surface represents the total amount of line that is near each location. The units of	
    the calculated density values are the length of line per unit area. For example, if the lines represent rivers, the	
    result layer will represent the total length of rivers that are within the search radius. This result can be used to	
    identify areas that are hospitable to grazing animals.
    
    =========================    =========================================================
    **Argument**                 **Description**
    -------------------------    ---------------------------------------------------------
    input_layer                  Required layer. The point or line features from which to calculate density. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    field                        Optional string. A numeric field name specifying the number of incidents at each location.  For example, if you have points that represent cities, you can use a field representing the population of the city as the count field, and the resulting population density layer will calculate larger population densities near cities with larger populations. If not specified, each location will be assumed to represent a single count.
    -------------------------    ---------------------------------------------------------
    cell_size                    Optional float. This value is used to create a mesh of points where density values are calculated. The default is approximately 1/1000th of the smaller of the width and height of the analysis extent as defined in the context parameter. The smaller the value, the smoother the polygon boundaries will be. Conversely, with larger values, the polygon boundaries will be more coarse and jagged.
    -------------------------    ---------------------------------------------------------
    cell_size_units              Optional string. The units of the cell_size value.
                                 Choice list: ['Miles', 'Feet', 'Kilometers',  'Meters']
    -------------------------    ---------------------------------------------------------
    radius                       Optional float. A distance specifying how far to search to find point or line features when calculating density values.
    -------------------------    ---------------------------------------------------------
    radius_units                 Optional string. The units of the radius parameter. If no distance is provided, a default will be calculated that is based on the locations of the input features and the values in the count field (if a count field is provided).
                                 Choice list: ['Miles', 'Feet', 'Kilometers',  'Meters']
    -------------------------    ---------------------------------------------------------
    bounding_polygon_layer       Optional layer. A layer specifying the polygon(s) where you want densities to be calculated. For example, if you are interpolating densities of fish within a lake, you can use the boundary of the lake in this parameter and the output will only draw within the boundary of the lake. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    area_units                   Optional string. The units of the calculated density values.
                                 Choice list: ['areaUnits', 'SquareMiles']
    -------------------------    ---------------------------------------------------------	
    classification_type          Optional string. Determines how density values will be classified into polygons.
                                 Choice list: ['EqualInterval', 'GeometricInterval', 'NaturalBreaks', 'EqualArea', 'StandardDeviation']	
                                    * EqualInterval—Polygons are created such that the range of density values is equal for each area.	
                                    * GeometricInterval—Polygons are based on class intervals that have a geometric series. This method ensures that each class range has approximately the same number of values within each class and that the change between intervals is consistent.	
                                    * NaturalBreaks—Class intervals for polygons are based on natural groupings of the data. Class break values are identified that best group similar values and that maximize the differences between classes.	
                                    * EqualArea—Polygons are created such that the size of each area is equal. For example, if the result has more high density values than low density values, more polygons will be created for high densities.	
                                    * StandardDeviation—Polygons are created based upon the standard deviation of the predicted density values.
    -------------------------    ---------------------------------------------------------
    num_classes                  Optional int. This value is used to divide the range of predicted values into distinct classes. The range of values in each class is determined by the classification_type parameter.
    -------------------------    ---------------------------------------------------------
    output_name                  Optional string. Additional properties such as output feature service name.	
    -------------------------    ---------------------------------------------------------
    context                      Optional string. Additional settings such as processing extent and output spatial reference. For calculate_density, there are two settings.

                                 #. Extent (extent)-a bounding box that defines the analysis area. Only those points in the input_layer that intersect the bounding box will be analyzed.	
                                 #. Output Spatial Reference (outSR)—the output features will be projected into the output spatial reference.
    -------------------------    ---------------------------------------------------------
    gis                          Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------    ---------------------------------------------------------
    estimate                     Optional Boolean. Is true, the number of credits needed to run the operation will be returned as a float.
    =========================    =========================================================


    :returns: result_layer : feature layer Item if output_name is specified, else Feature Collection.

    .. code-block:: python

        USAGE EXAMPLE: To create a layer that shows density of collisions within 2 miles.	
                       The density is classified based upon the standard deviation.
                       The range of density values is divided into 5 classes.	
        	
        collision_density = calculate_density(input_layer=collisions,	
                                        radius=2,	
                                        radius_units='Miles',	
                                        bounding_polygon_layer=zoning_lyr,	
                                        area_units='SquareMiles',	
                                        classification_type='StandardDeviation',	
                                        num_classes=5,	
                                        output_name='density_of_incidents')

    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.calculate_density(
        input_layer,
        field,
        cell_size,
        cell_size_units,
        radius,
        radius_units,
        bounding_polygon_layer,
        area_units,
        classification_type,
        num_classes,
        output_name,
        context,
        estimate=estimate)

def summarize_center_and_dispersion(
        analysis_layer,
        summarize_type,
        ellipse_size=None,
        weight_field=None,
        group_field=None,
        output_name=None,
        context=None,
        gis=None,
        estimate=False):

    """
    The Summarize Center and Dispersion task finds central features and directional distributions.

    ====================    =========================================================
    **Argument**            **Description**
    --------------------    ---------------------------------------------------------
    analysis_layer          The point, line, or polygon features to be analyzed. This
                            parameter can be a URL to a feature service layer with an
                            optional filter to select specific feaures, or a feature
                            collection
    --------------------    ---------------------------------------------------------
    summarize_type          The method with which to summarize the analysis_layer.
                            Choice List:
                            ["CentralFeature", "MeanCenter", "MedianCenter",
                            "Ellipse"]
                            Example: "CentralFeature"
    --------------------    ---------------------------------------------------------
    ellipse_size            The size of the output ellipse in standard deviations.
                            The default ellipse size is 1. Valid choices are 1, 2, or
                            3 standard deviations.
                            Choice List: [1, 2, 3]
                            Examples:
                            "1"
                            [1, 2, 3]
    --------------------    ---------------------------------------------------------
    weight_field            A numeric field in the analysis_layer to be used to
                            weight locations according to their relative importance.
    --------------------    ---------------------------------------------------------
    group_field             The field used to group features for separate directional
                            distribution calculations. The group_field can be of
                            integer, date, or string type.
    --------------------    ---------------------------------------------------------
    output_name             Optional string. Additional properties such as output
                            feature service name.
    --------------------    ---------------------------------------------------------
    context                 Optional string. Additional settings such as processing
                            extent and output spatial reference.
    --------------------    ---------------------------------------------------------
    gis                     Optional, the GIS on which this tool runs. If not
                            specified, the active GIS is used.
    --------------------    ---------------------------------------------------------
    estimate                Optional Boolean. If True, the number of credits to run the operation will be returned.
    ====================    =========================================================

    :returns: Python dictionary with the following keys:
        "central_feature_result_layer" : layer (FeatureCollection)
        "mean_feature_result_layer" : layer (FeatureCollection)
        "median_feature_result_layer" : layer (FeatureCollection)
        "ellipse_feature_result_layer" : layer (FeatureCollection)
        "process_info" : list of messages
    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.summarize_center_and_dispersion(
        analysis_layer,
        summarize_type,
        ellipse_size,
        weight_field,
        group_field,
        output_name,
        context,
        estimate=estimate)

def find_point_clusters(
        analysis_layer,
        min_features_cluster,
        search_distance=None,
        search_distance_unit=None,
        output_name=None,
        context=None,
        gis=None, estimate=False):

    """
    The Find Point Clusters function finds clusters of point features in surrounding
    noise based on their spatial distribution. Output is a layer containing records
    assigned to a cluster or noise.

    ====================    =========================================================
    **Argument**            **Description**
    --------------------    ---------------------------------------------------------
    analysis_layer          Required layer. The point feature layer for which
                            density-based clustering will be calculated.
    --------------------    ---------------------------------------------------------
    min_features_cluster    Required integer. The minimum number of features to be
                            considered a cluster. Any cluster with fewer features
                            than the number provided will be considered noise.
    --------------------    ---------------------------------------------------------
    search_distance         Optional double. The maximum distance to consider. The
                            Minimum Features per Cluster specified must be found
                            within this distance for cluster membership. Individual
                            clusters will be separated by at least this distance. If
                            a feature is located further than this distance from the
                            next closest feature in the cluster, it will not be
                            included in the cluster.
    --------------------    ---------------------------------------------------------
    search_distance_unit    Optional string. The linear unit to be used for the
                            search distance parameter.
    --------------------    ---------------------------------------------------------
    output_name             Optional string. Additional properties such as output
                            feature service name.
    --------------------    ---------------------------------------------------------
    context                 Optional string. Additional settings such as processing
                            extent and output spatial reference.
    --------------------    ---------------------------------------------------------
    gis                     Optional, the GIS on which this tool runs. If not
                            specified, the active GIS is used.
    --------------------    ---------------------------------------------------------
    estimate                Optional Boolean. If True, the number of credits to run the operation will be returned.
    ====================    =========================================================

    :returns: Python dictionary with the following keys:
        "point_clusters_result_layer" : layer (FeatureCollection)
        "process_info" : list of messages
    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_point_clusters(
        analysis_layer,
        min_features_cluster,
        search_distance,
        search_distance_unit,
        output_name,
        context,
        estimate=estimate)

def find_hot_spots(
        analysis_layer,
        analysis_field=None,
        divided_by_field=None,
        bounding_polygon_layer=None,
        aggregation_polygon_layer=None,
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        shape_type=None):
    """
    The Find Hot Spots function finds statistically significant clusters of incident points, weighted points, or
    weighted polygons. For incident data, the analysis field (weight) is obtained by aggregation.
    Output is a hot spot map.

    =========================    =========================================================
    **Argument**                 **Description**
    -------------------------    ---------------------------------------------------------
    analysis_layer               Required layer (see Feature Input in documentation). The point or polygon feature layer for which hot spots will be calculated.
    -------------------------    ---------------------------------------------------------
    analysis_field               Optional string. The numeric field in the AnalysisLayer that will be analyzed.
    -------------------------    ---------------------------------------------------------
    divided_by_field             Optional string. The field that will segment the locations.
    -------------------------    ---------------------------------------------------------
    bounding_polygon_layer       Optional layer (see Feature Input in documentation). When the analysis layer is points and no AnalysisField is specified, you can provide polygons features that define where incidents could have occurred.
    -------------------------    ---------------------------------------------------------
    aggregation_polygon_layer    Optional layer (see Feature Input in documentation). When the AnalysisLayer contains points and no AnalysisField is specified, you can provide polygon features into which the points will be aggregated and analyzed, such as administrative units.
    -------------------------    ---------------------------------------------------------
    output_name                  Optional string. Additional properties such as output feature service name.
    -------------------------    ---------------------------------------------------------
    context                      Optional string. Additional settings such as processing extent and output spatial reference.
    -------------------------    ---------------------------------------------------------
    gis                          Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------    ---------------------------------------------------------
    estimate                     Optional Boolean. Is true, the number of credits needed to run the operation will be returned as a float.
    -------------------------    ---------------------------------------------------------
    shape_type                   Optional string. The shape of the polygon mesh the input features will be aggregated into.
    =========================    =========================================================

    :Returns: dict with the following keys:
       "hot_spots_result_layer" : layer (FeatureCollection)
       "process_info" : list of messages

    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_hot_spots(
        analysis_layer,
        analysis_field,
        divided_by_field,
        bounding_polygon_layer,
        aggregation_polygon_layer,
        output_name,
        context,
        estimate=estimate,
        shape_type=shape_type)


def find_outliers(analysis_layer,
                  analysis_field,
                  divided_by_field=None,
                  bounding_polygon_layer=None,
                  aggregation_polygon_layer=None,
                  permutations=None,
                  shape_type=None,
                  cell_size=None,
                  cell_units=None,
                  distance_band=None,
                  band_units=None,
                  output_name=None,
                  context=None,
                  gis=None,
                  estimate=False):
    """

    The Find Outliers task analyzes point data (such as crime incidents, traffic accidents, or trees) or field values associated with points or area features (such as the number of people in each census tract or the total sales for retail stores). It finds statistically significant spatial clusters of high values and low values and statistically significant high or low spatial outliers within those clusters.

    The result map layer shows high outliers in red and low outliers in dark blue. Clusters of high values appear pink and clusters of low values appear light blue. Features that are beige are not a statistically significant outlier and not part of a statistically significant cluster; the spatial pattern associated with these features could very likely be the result of random processes and random chance.

    Parameters
    ----------
    analysis_layer : Required layer (see Feature Input in documentation)
        The point or polygon feature layer for which outliers will be calculated.
    analysis_field : Optional string
        The numeric field that will be analyzed.
    divided_by_field : Optional string, The numeric field in the analysis_layer that will be used to normalize your data.
    bounding_polygon_layer : Optional layer (see Feature Input in documentation)
        When the analysis layer is points and no analysisField is specified, you can provide polygon features that define where incidents could have occurred.
    aggregation_polygon_layer : Optional layer (see Feature Input in documentation)
        When the AnalysisLayer contains points and no AnalysisField is specified, you can provide polygon features into which the points will be aggregated and analyzed, such as administrative units.
    permutations : Permutations are used to determine how likely it would be to find the actual spatial distribution of the values you are analyzing. Choosing the number of permutations is a balance between precision and increased processing time. A lower number of permutations can be used when first exploring a problem, but it is best practice to increase the permutations to the highest number feasible for final results.

       - Speed implements 199 permutations and results in p-values with a precision of 0.01.
       - Balance implements 499 permutations and results in p-values with a precision of 0.002.
       - Precision implements 999 permutations and results in p-values with a precision of 0.001.
       Values: Speed | Balance | Precision
    shape_type : optional string, The shape of the polygon mesh the input features will be aggregated into.

      - Fishnet - The input features will be aggregated into a grid of square (fishnet) cells.
      - Hexagon - The input features will be aggregated into a grid of hexagonal cells.
    cell_size : The size of the grid cells used to aggregate your features. When aggregating into a hexagon grid, this distance is used as the height to construct the hexagon polygons.
    cell_units : The units of the cellSize value. You must provide a value if cellSize has been set.
      Values: Miles | Feet | Kilometers | Meters
    distance_band : The spatial extent of the analysis neighborhood. This value determines which features are analyzed together in order to assess local clustering.
    band_units : The units of the distanceBand value. You must provide a value if distanceBand has been set.
      Values: Miles | Feet | Kilometers | Meters
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis : The GIS used for running this analysis
    estimate : Optional Boolean. If True, the number of credits to run the operation will be returned.

    Returns
    -------
    Item it output_name is set.
    dict with the following keys:
       "find_outliers_result_layer" : layer (FeatureCollection)
       "process_info" : list of messages

        """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_outliers(analysis_layer,
                                                    analysis_field,
                                                    divided_by_field,
                                                    bounding_polygon_layer,
                                                    aggregation_polygon_layer,
                                                    permutations,
                                                    shape_type,
                                                    cell_size,
                                                    cell_units,
                                                    distance_band,
                                                    band_units,
                                                    output_name,
                                                    context,
                                                    estimate=estimate)


def interpolate_points(
        input_layer,
        field,
        interpolate_option="5",
        output_prediction_error=False,
        classification_type="GeometricInterval",
        num_classes=10,
        class_breaks=[],
        bounding_polygon_layer=None,
        predict_at_point_layer=None,
        output_name=None,
        context=None,
        gis=None,
        estimate=False):
    """
    The Interpolate Points function allows you to predict values at new locations based on measurements from a
    collection of points. The function takes point data with values at each point and returns areas classified by
    predicted values.

    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)
        The point layer whose features will be interpolated.
    field : Required string
        Name of the numeric field containing the values you wish to interpolate.
    interpolate_option : Optional string
        Integer value declaring your preference for speed versus accuracy, from 1 (fastest) to 9 (most accurate). More
        accurate predictions take longer to calculate.
    output_prediction_error : Optional bool
        If True, a polygon layer of standard errors for the interpolation predictions will be returned in the
        predictionError output parameter.
    classification_type : Optional string
        Determines how predicted values will be classified into areas.
    num_classes : Optional int
        This value is used to divide the range of interpolated values into distinct classes. The range of values in each
        class is determined by the classificationType parameter. Each class defines the boundaries of the result
        polygons.
    class_breaks : Optional list of floats
        If classificationType is Manual, supply desired class break values separated by spaces. These values define the
        upper limit of each class, so the number of classes will equal the number of entered values. Areas will not be
        created for any locations with predicted values above the largest entered break value. You must enter at least
        two values and no more than 32.
    bounding_polygon_layer : Optional layer (see Feature Input in documentation)
        A layer specifying the polygon(s) where you want values to be interpolated.
    predict_at_point_layer : Optional layer (see Feature Input in documentation)
        An optional layer specifying point locations to calculate prediction values. This allows you to make predictions
        at specific locations of interest.
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    estimate :
        Optional Boolean. If True, the number of credits to run the operation will be returned.

    Returns
    -------
    dict with the following keys:
       "result_layer" : layer (FeatureCollection)
       "prediction_error" : layer (FeatureCollection)
       "predicted_point_layer" : layer (FeatureCollection)
    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.interpolate_points(
        input_layer,
        field,
        interpolate_option,
        output_prediction_error,
        classification_type,
        num_classes,
        class_breaks,
        bounding_polygon_layer,
        predict_at_point_layer,
        output_name,
        context,
        estimate=estimate)



try:
    import torch
    from torchvision import models
    from fastai.vision.learner import unet_learner
    import numpy as np
    import json
    import os
    from pathlib import Path
    from ._codetemplate import image_classifier_prf
    from ._ssd import _raise_fastai_import_error, _EmptyData
    from functools import partial
    from ._unet_utils import is_no_color, LabelCallback
    HAS_FASTAI = True
except Exception as e:
    HAS_FASTAI = False


class ArcGISModel(object):
    
    def lr_find(self):
        """
        Runs the Learning Rate Finder, and displays the graph of it's output.
        Helps in choosing the optimum learning rate for training the model.
        """
        from IPython.display import clear_output
        self.learn.lr_find()
        clear_output()
        self.learn.recorder.plot()
    
    
    def fit(self, epochs=10, lr=slice(1e-4,3e-3), one_cycle=True):
        """
        Train the model for the specified number of epocs and using the
        specified learning rates
        
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        epochs                  Required integer. Number of cycles of training
                                on the data. Increase it if underfitting.
        ---------------------   -------------------------------------------
        lr                      Required float or slice of floats. Learning rate
                                to be used for training the model. Select from
                                the `lr_find` plot.
        ---------------------   -------------------------------------------
        one_cycle               Optional boolean. Parameter to select 1cycle
                                learning rate schedule. If set to `False` no 
                                learning rate schedule is used.                                
        =====================   ===========================================
        """
        if one_cycle:
            self.learn.fit_one_cycle(epochs, lr)
        else:
            self.learn.fit(epochs, lr)
        
    def unfreeze(self):
        """
        Unfreezes the earlier layers of the detector for fine-tuning.
        """
        self.learn.unfreeze()
        
    def save(self, name_or_path):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment to Image Server or ArcGIS Pro
        Train the model for the specified number of epocs and using the
        specified learning rates.
        
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name of the model to save. It
                                stores it at the pre-defined location. If path
                                is passed then it stores at the specified path
                                with model name as directory name. and creates
                                all the intermediate directories.
        =====================   ===========================================
        """
        if '\\' in name_or_path or '/' in name_or_path:
            path = Path(name_or_path)
            name = path.parts[-1]
            # to make fastai save to both path and with name
            temp = self.learn.path
            self.learn.path = path
            self.learn.model_dir = ''
            if not os.path.exists(self.learn.path):
                os.makedirs(self.learn.path)
            saved_path = self.learn.save(name, return_path=True)
            # undoing changes to self.learn.path and self.learn.model
            self.learn.path = temp
            self.learn.model_dir = 'models'
        else:
            temp = self.learn.path
            # fixing fastai bug
            self.learn.path = self.learn.path.parent
            self.learn.model_dir =  Path(self.learn.model_dir) /  name_or_path
            if not os.path.exists(self.learn.path / self.learn.model_dir):
                os.makedirs(self.learn.path / self.learn.model_dir)
            saved_path = self.learn.save(name_or_path,  return_path=True)
            # undoing changes to self.learn.path
            self.learn.path = temp
            self.learn.model_dir = 'models'

        zip_name = self._create_emd(saved_path)
        with open(saved_path.parent / self._emd_template['InferenceFunction'], 'w') as f:
            f.write(self._code)
        self._create_zip(zip_name, str(saved_path.parent))
        print('Created model files at {spp}'.format(spp=saved_path.parent))
        
    def _create_zip(self, zipname, path):
        import shutil
        zip_file = shutil.make_archive(zipname, 'zip', path)
        shutil.move(zip_file, path)
        
    def load(self, name_or_path):
        """
        Loads a saved model for inferencing or fine tuning from the specified
        path or model name.
        
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name of the model to load from
                                the pre-defined location. If path is passed then
                                it loads from the specified path with model name
                                as directory name. Path to ".pth" file can also
                                be passed
        =====================   ===========================================
        """
        if '\\' in name_or_path or '/' in name_or_path:
            path = Path(name_or_path)
            # to make fastai from both path and with name
            temp = self.learn.path
            if path.is_file():
                name = path.stem
                self.learn.path = path.parent
            else:
                name = path.parts[-1]
                self.learn.path = path
            self.learn.model_dir = ''
            self.learn.load(name)
            # undoing changes to self.learn.path and self.learn.model_dir
            self.learn.path = temp
            self.learn.model_dir = 'models'
        else:
            temp = self.learn.path
            # fixing fastai bug
            self.learn.path = self.learn.path.parent
            self.learn.model_dir =  Path(self.learn.model_dir) /  name_or_path
            self.learn.load(name_or_path)
            # undoing changes to self.learn.path
            self.learn.path = temp
            self.learn.model_dir = 'models'  

_CLASS_TEMPLATE =     {
      "Value" : 1,
      "Name" : "1",
      "Color" : []
    }

_EMD_TEMPLATE = {
    "Framework":"arcgis.learn.models._inferencing",
    "ModelConfiguration":"_unet",
    "ModelFile":"",
    "InferenceFunction": "ArcGISImageClassifier.py",
    "ExtractBands":[0,1,2],
    "ImageWidth":400,
    "ImageHeight":400,
    "Classes" : []
}

def accuracy(input, target, void_code=0, class_mapping=None):  
    target = target.squeeze(1) 
    mask = target != void_code
    return (input.argmax(dim=1)[mask] == target[mask]).float().mean()

class UnetClassifier(ArcGISModel):

    """
    Creates a Unet like classifier based on given pretrained encoder.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the `UnetClassifier`, which
                            is `resnet34` by default.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    :returns: `UnetClassifier` Object
    """

    
    def __init__(self, data, backbone=None, pretrained_path=None):
      
        super().__init__()

        self._device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

        if not HAS_FASTAI:
            _raise_fastai_import_error()

        self._emd_template = _EMD_TEMPLATE
        self._code = image_classifier_prf

        if backbone is None:
            self._backbone = models.resnet34
        elif type(backbone) is str:
            self._backbone = getattr(models, backbone)
        else:
            self._backbone = backbone

        acc_metric = partial(accuracy, void_code=0, class_mapping=data.class_mapping)
        self._data = data
        self.learn = unet_learner(data, arch=self._backbone, metrics=acc_metric, wd=1e-2, bottle=True, last_cross=True)
        self.learn.model = self.learn.model.to(self._device)

        if pretrained_path is not None:
            self.load(pretrained_path)
        
    @classmethod
    def from_emd(cls, data, emd_path):
        emd_path = Path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)
            
        model_file = Path(emd['ModelFile'])
        
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file
            
        model_params = emd['ModelParameters']

        try:
            class_mapping = {i['Value'] : i['Name'] for i in emd['Classes']}
            color_mapping = {i['Value'] : i['Color'] for i in emd['Classes']}
        except KeyError:
            class_mapping = {i['ClassValue'] : i['ClassName'] for i in emd['Classes']} 
            color_mapping = {i['ClassValue'] : i['Color'] for i in emd['Classes']}                

        
        if data is None:
            empty_data = _EmptyData(path='str', loss_func=None, c=len(class_mapping) + 1, chip_size=emd['ImageHeight'])
            empty_data.class_mapping = class_mapping
            empty_data.color_mapping = color_mapping
            return cls(empty_data, **model_params, pretrained_path=str(model_file))
        else:
            return cls(data, **model_params, pretrained_path=str(model_file))        
        
        
    def _create_emd(self, path):
        import random
        _EMD_TEMPLATE['ModelFile'] = path.name
        _EMD_TEMPLATE['ImageHeight'] = self._data.chip_size
        _EMD_TEMPLATE['ImageWidth'] = self._data.chip_size
        _EMD_TEMPLATE['ModelParameters'] = {
                                            'backbone': self._backbone.__name__
                                           }
        _EMD_TEMPLATE['Classes'] = []
        for i, class_name in enumerate(self._data.classes[1:]): # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            _CLASS_TEMPLATE["Value"] = inverse_class_mapping[class_name]
            _CLASS_TEMPLATE["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)] if is_no_color(self._data.color_mapping) else self._data.color_mapping[inverse_class_mapping[class_name]]
            _CLASS_TEMPLATE["Color"] = color
            _EMD_TEMPLATE['Classes'].append(_CLASS_TEMPLATE.copy())
            
        json.dump(_EMD_TEMPLATE, open(path.with_suffix('.emd'), 'w'), indent=4)
        return path.stem

    def fit(self, epochs=10, lr=slice(1e-4,3e-3), one_cycle=True):
        """
        Train the model for the specified number of epocs and using the
        specified learning rates
        
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        epochs                  Required integer. Number of cycles of training
                                on the data. Increase it if underfitting.
        ---------------------   -------------------------------------------
        lr                      Required float or slice of floats. Learning rate
                                to be used for training the model. Select from
                                the `lr_find` plot.
        ---------------------   -------------------------------------------
        one_cycle               Optional boolean. Parameter to select 1cycle
                                learning rate schedule. If set to `False` no 
                                learning rate schedule is used.                                
        =====================   ===========================================
        """
        if one_cycle:
            self.learn.fit_one_cycle(epochs, lr, callbacks=[LabelCallback(self.learn)])
        else:
            self.learn.fit(epochs, lr, callbacks=[LabelCallback(self.learn)])


    def show_results(self, rows=5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        """
        if rows > self._data.batch_size:
            rows = self._data.batch_size
        self.learn.show_results(rows=rows, **kwargs)             
